# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import functools
import random
from subprocess import Popen
from typing import Tuple

import dateutil
from kafka import KafkaProducer

from swh.storage import get_storage

from swh.journal.client import JournalClient
from swh.journal.serializers import key_to_kafka, value_to_kafka
from swh.journal.replay import process_replay_objects

from .conftest import OBJECT_TYPE_KEYS


def test_storage_play(
        kafka_prefix: str,
        kafka_server: Tuple[Popen, int]):
    (_, port) = kafka_server
    kafka_prefix += '.swh.journal.objects'

    storage = get_storage('memory', {})

    producer = KafkaProducer(
        bootstrap_servers='localhost:{}'.format(port),
        key_serializer=key_to_kafka,
        value_serializer=value_to_kafka,
        client_id='test producer',
    )

    now = datetime.datetime.now(tz=datetime.timezone.utc)

    # Fill Kafka
    nb_sent = 0
    nb_visits = 0
    for (object_type, (_, objects)) in OBJECT_TYPE_KEYS.items():
        topic = kafka_prefix + '.' + object_type
        for object_ in objects:
            key = bytes(random.randint(0, 255) for _ in range(40))
            object_ = object_.copy()
            if object_type == 'content':
                object_['ctime'] = now
            elif object_type == 'origin_visit':
                nb_visits += 1
                object_['visit'] = nb_visits
            producer.send(topic, key=key, value=object_)
            nb_sent += 1

    # Fill the storage from Kafka
    config = {
        'brokers': 'localhost:%d' % kafka_server[1],
        'group_id': 'replayer',
        'prefix': kafka_prefix,
        'max_messages': nb_sent,
    }
    replayer = JournalClient(**config)
    worker_fn = functools.partial(process_replay_objects, storage=storage)
    nb_inserted = 0
    while nb_inserted < nb_sent:
        nb_inserted += replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # Check the objects were actually inserted in the storage
    assert OBJECT_TYPE_KEYS['revision'][1] == \
        list(storage.revision_get(
            [rev['id'] for rev in OBJECT_TYPE_KEYS['revision'][1]]))
    assert OBJECT_TYPE_KEYS['release'][1] == \
        list(storage.release_get(
            [rel['id'] for rel in OBJECT_TYPE_KEYS['release'][1]]))

    origins = list(storage.origin_get(
            [orig for orig in OBJECT_TYPE_KEYS['origin'][1]]))
    assert OBJECT_TYPE_KEYS['origin'][1] == \
        [{'url': orig['url'], 'type': orig['type']} for orig in origins]
    for origin in origins:
        expected_visits = [
            {
                **visit,
                'origin': origin['id'],
                'date': dateutil.parser.parse(visit['date']),
            }
            for visit in OBJECT_TYPE_KEYS['origin_visit'][1]
            if visit['origin']['url'] == origin['url']
            and visit['origin']['type'] == origin['type']
        ]
        actual_visits = list(storage.origin_visit_get(origin['id']))
        for visit in actual_visits:
            del visit['visit']  # opaque identifier
        assert expected_visits == actual_visits

    contents = list(storage.content_get_metadata(
            [cont['sha1'] for cont in OBJECT_TYPE_KEYS['content'][1]]))
    assert None not in contents
    assert contents == OBJECT_TYPE_KEYS['content'][1]



# Copyright (C) 2017 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import OrderedDict
import itertools
import unittest

from swh.journal import serializers


class TestSerializers(unittest.TestCase):
    def test_key_to_kafka_repeatable(self):
        """Check the kafka key encoding is repeatable"""
        base_dict = {
            'a': 'foo',
            'b': 'bar',
            'c': 'baz',
        }

        key = serializers.key_to_kafka(base_dict)

        for dict_keys in itertools.permutations(base_dict):
            d = OrderedDict()
            for k in dict_keys:
                d[k] = base_dict[k]

            self.assertEqual(key, serializers.key_to_kafka(d))



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# the default prefix for kafka's topics
DEFAULT_PREFIX = 'swh.journal.objects'



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from swh.storage import HashCollision
from swh.model.hashutil import hash_to_hex
from swh.objstorage.objstorage import ID_HASH_ALGO


logger = logging.getLogger(__name__)


def process_replay_objects(all_objects, *, storage):
    for (object_type, objects) in all_objects.items():
        _insert_objects(object_type, objects, storage)


def _insert_objects(object_type, objects, storage):
    if object_type == 'content':
        # TODO: insert 'content' in batches
        for object_ in objects:
            try:
                storage.content_add_metadata([object_])
            except HashCollision as e:
                logger.error('Hash collision: %s', e.args)
    elif object_type in ('directory', 'revision', 'release',
                         'snapshot', 'origin'):
        # TODO: split batches that are too large for the storage
        # to handle?
        method = getattr(storage, object_type + '_add')
        method(objects)
    elif object_type == 'origin_visit':
        storage.origin_visit_upsert([
            {
                **obj,
                'origin': storage.origin_add_one(obj['origin'])
            }
            for obj in objects])
    else:
        logger.warning('Received a series of %s, this should not happen',
                       object_type)


def process_replay_objects_content(all_objects, *, src, dst):
    for (object_type, objects) in all_objects.items():
        if object_type != 'content':
            logger.warning('Received a series of %s, this should not happen',
                           object_type)
            continue
        logger.info('processing %s content objects', len(objects))
        for obj in objects:
            obj_id = obj[ID_HASH_ALGO]
            if obj['status'] == 'visible':
                try:
                    obj = src.get(obj_id)
                    dst.add(obj, obj_id=obj_id)
                    logger.debug('copied %s', hash_to_hex(obj_id))
                except Exception:
                    logger.exception('Failed to copy %s', hash_to_hex(obj_id))
            else:
                logger.debug('skipped %s (%s)',
                             hash_to_hex(obj_id), obj['status'])



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.journal.backfill import (
    JournalBackfiller, compute_query, PARTITION_KEY
)


TEST_CONFIG = {
    'brokers': ['localhost'],
    'prefix': 'swh.tmp_journal.new',
    'client_id': 'swh.journal.client.test',
    'storage_dbconn': 'service=swh-dev',
}


def test_config_ko_missing_mandatory_key():
    """Missing configuration key will make the initialization fail

    """
    for key in TEST_CONFIG.keys():
        config = TEST_CONFIG.copy()
        config.pop(key)

        with pytest.raises(ValueError) as e:
            JournalBackfiller(config)

        error = ('Configuration error: The following keys must be'
                 ' provided: %s' % (','.join([key]), ))
        assert e.value.args[0] == error


def test_config_ko_unknown_object_type():
    """Parse arguments will fail if the object type is unknown

    """
    backfiller = JournalBackfiller(TEST_CONFIG)
    with pytest.raises(ValueError) as e:
        backfiller.parse_arguments('unknown-object-type', 1, 2)

    error = ('Object type unknown-object-type is not supported. '
             'The only possible values are %s' % (
                 ', '.join(PARTITION_KEY)))
    assert e.value.args[0] == error


def test_compute_query_content():
    query, where_args, column_aliases = compute_query(
        'content', '\x000000', '\x000001')

    assert where_args == ['\x000000', '\x000001']

    assert column_aliases == [
        'sha1', 'sha1_git', 'sha256', 'blake2s256', 'length', 'status',
        'ctime'
    ]

    assert query == '''
select sha1,sha1_git,sha256,blake2s256,length,status,ctime
from content

where (sha1) >= %s and (sha1) < %s
    '''


def test_compute_query_skipped_content():
    query, where_args, column_aliases = compute_query(
        'skipped_content', None, None)

    assert where_args == []

    assert column_aliases == [
        'sha1', 'sha1_git', 'sha256', 'blake2s256', 'length', 'ctime',
        'status', 'reason',
    ]

    assert query == '''
select sha1,sha1_git,sha256,blake2s256,length,ctime,status,reason
from skipped_content


    '''


def test_compute_query_origin_visit():
    query, where_args, column_aliases = compute_query(
        'origin_visit', 1, 10)

    assert where_args == [1, 10]

    assert column_aliases == [
        'type', 'url', 'date', 'snapshot', 'status'
    ]

    assert query == '''
select type,url,date,snapshot,status
from origin_visit
left join origin on origin_visit.origin=origin.id
where (origin_visit.origin) >= %s and (origin_visit.origin) < %s
    '''


def test_compute_query_release():
    query, where_args, column_aliases = compute_query(
        'release', '\x000002', '\x000003')

    assert where_args == ['\x000002', '\x000003']

    assert column_aliases == [
        'id', 'date', 'date_offset', 'comment', 'name', 'synthetic',
        'date_neg_utc_offset', 'target', 'target_type', 'author_id',
        'author_name', 'author_email', 'author_fullname']

    assert query == '''
select release.id as id,date,date_offset,comment,release.name as name,synthetic,date_neg_utc_offset,target,target_type,a.id as author_id,a.name as author_name,a.email as author_email,a.fullname as author_fullname
from release
left join person a on release.author=a.id
where (release.id) >= %s and (release.id) < %s
    '''  # noqa



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re
import tempfile
from subprocess import Popen
from typing import Tuple
from unittest.mock import patch

from click.testing import CliRunner
from kafka import KafkaProducer
import pytest

from swh.storage.in_memory import Storage

from swh.journal.cli import cli
from swh.journal.serializers import key_to_kafka, value_to_kafka


CLI_CONFIG = '''
storage:
    cls: memory
    args: {}
'''


@pytest.fixture
def storage():
    """An instance of swh.storage.in_memory.Storage that gets injected
    into the CLI functions."""
    storage = Storage()
    with patch('swh.journal.cli.get_storage') as get_storage_mock:
        get_storage_mock.return_value = storage
        yield storage


def invoke(catch_exceptions, args):
    runner = CliRunner()
    with tempfile.NamedTemporaryFile('a', suffix='.yml') as config_fd:
        config_fd.write(CLI_CONFIG)
        config_fd.seek(0)
        args = ['-C' + config_fd.name] + args
        result = runner.invoke(cli, args)
    if not catch_exceptions and result.exception:
        print(result.output)
        raise result.exception
    return result


def test_replay(
        storage: Storage,
        kafka_prefix: str,
        kafka_server: Tuple[Popen, int]):
    (_, port) = kafka_server
    kafka_prefix += '.swh.journal.objects'

    producer = KafkaProducer(
        bootstrap_servers='localhost:{}'.format(port),
        key_serializer=key_to_kafka,
        value_serializer=value_to_kafka,
        client_id='test-producer',
    )

    snapshot = {'id': b'foo', 'branches': {
        b'HEAD': {
            'target_type': 'revision',
            'target': b'bar',
        }
    }}
    producer.send(
        topic=kafka_prefix+'.snapshot', key=snapshot['id'], value=snapshot)

    result = invoke(False, [
        'replay',
        '--broker', 'localhost:%d' % port,
        '--group-id', 'test-cli-consumer',
        '--prefix', kafka_prefix,
        '--max-messages', '1',
    ])
    expected = r'Done.\n'
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    assert storage.snapshot_get(snapshot['id']) == {
        **snapshot, 'next_branch': None}


# TODO: write a test for the content-replay command



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import namedtuple
import functools

from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import lists

from swh.model.hypothesis_strategies import object_dicts
from swh.storage.in_memory import Storage
from swh.storage import HashCollision

from swh.journal.client import JournalClient, ACCEPTED_OBJECT_TYPES
from swh.journal.direct_writer import DirectKafkaWriter
from swh.journal.replay import process_replay_objects
from swh.journal.replay import process_replay_objects_content
from swh.journal.serializers import (
    key_to_kafka, kafka_to_key, value_to_kafka, kafka_to_value)

FakeKafkaMessage = namedtuple('FakeKafkaMessage', 'key value')
FakeKafkaPartition = namedtuple('FakeKafkaPartition', 'topic')


class MockedKafkaWriter(DirectKafkaWriter):
    def __init__(self, queue):
        self._prefix = 'prefix'
        self.queue = queue

    def send(self, topic, key, value):
        key = kafka_to_key(key_to_kafka(key))
        value = kafka_to_value(value_to_kafka(value))
        partition = FakeKafkaPartition(topic)
        msg = FakeKafkaMessage(key=key, value=value)
        if self.queue and {partition} == set(self.queue[-1]):
            # The last message is of the same object type, groupping them
            self.queue[-1][partition].append(msg)
        else:
            self.queue.append({partition: [msg]})


class MockedKafkaConsumer:
    def __init__(self, queue):
        self.queue = queue
        self.committed = False

    def poll(self):
        return self.queue.pop(0)

    def commit(self):
        if self.queue == []:
            self.committed = True


class MockedJournalClient(JournalClient):
    def __init__(self, queue, object_types=ACCEPTED_OBJECT_TYPES):
        self._object_types = object_types
        self.consumer = MockedKafkaConsumer(queue)


@given(lists(object_dicts(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_write_replay_same_order_batches(objects):
    queue = []
    replayer = MockedJournalClient(queue)

    storage1 = Storage()
    storage1.journal_writer = MockedKafkaWriter(queue)

    for (obj_type, obj) in objects:
        obj = obj.copy()
        if obj_type == 'origin_visit':
            origin_id = storage1.origin_add_one(obj.pop('origin'))
            if 'visit' in obj:
                del obj['visit']
            storage1.origin_visit_add(origin_id, **obj)
        else:
            method = getattr(storage1, obj_type + '_add')
            try:
                method([obj])
            except HashCollision:
                pass

    queue_size = sum(len(partition)
                     for batch in queue
                     for partition in batch.values())

    storage2 = Storage()
    worker_fn = functools.partial(process_replay_objects, storage=storage2)
    nb_messages = 0
    while nb_messages < queue_size:
        nb_messages += replayer.process(worker_fn)

    assert replayer.consumer.committed

    for attr_name in ('_contents', '_directories', '_revisions', '_releases',
                      '_snapshots', '_origin_visits', '_origins'):
        assert getattr(storage1, attr_name) == getattr(storage2, attr_name), \
            attr_name


# TODO: add test for hash collision


@given(lists(object_dicts(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_write_replay_content(objects):
    queue = []
    replayer = MockedJournalClient(queue)

    storage1 = Storage()
    storage1.journal_writer = MockedKafkaWriter(queue)

    for (obj_type, obj) in objects:
        obj = obj.copy()
        if obj_type == 'content':
            storage1.content_add([obj])

    queue_size = sum(len(partition)
                     for batch in queue
                     for partition in batch.values())

    storage2 = Storage()
    worker_fn = functools.partial(process_replay_objects_content,
                                  src=storage1.objstorage,
                                  dst=storage2.objstorage)
    nb_messages = 0
    while nb_messages < queue_size:
        nb_messages += replayer.process(worker_fn)

    assert storage1.objstorage.state == storage2.objstorage.state
